import re
import typing as T
from datetime import timedelta
from pathlib import Path
from uuid import UUID
import pydantic
import traceback
import json

from mirai.event.message.models import FriendMessage, GroupMessage, BotMessage, MessageTypes

from mirai.event import ExternalEvent
from mirai.entities.friend import Friend
from mirai.entities.group import Group, GroupSetting, Member, MemberChangeableSetting
from mirai.event.message.chain import MessageChain
from mirai.misc import ImageRegex, ImageType, assertOperatorSuccess, raiser, printer, getMatchedString
from mirai.network import fetch
from mirai.event.message.base import BaseMessageComponent
from mirai.event.message import components
from mirai.misc import protocol_log, edge_case_handler
from mirai.image import InternalImage
from mirai.logger import Protocol as ProtocolLogger
import threading

class MiraiProtocol:
    qq: int
    baseurl: str
    session_key: str
    auth_key: str

    @protocol_log
    @edge_case_handler
    async def auth(self):
        return assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/auth", {
                "authKey": self.auth_key
            }
        ), raise_exception=True, return_as_is=True)

    @protocol_log
    @edge_case_handler
    async def verify(self):
        return assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/verify", {
                "sessionKey": self.session_key,
                "qq": self.qq
            }
        ), raise_exception=True, return_as_is=True)

    @protocol_log
    @edge_case_handler
    async def release(self):
        return assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/release", {
                "sessionKey": self.session_key,
                "qq": self.qq
            }
        ), raise_exception=True)

    @edge_case_handler
    async def getConfig(self) -> dict:
        return assertOperatorSuccess(
            await fetch.http_get(f"{self.baseurl}/config", {
                "sessionKey": self.session_key
            }
        ), raise_exception=True, return_as_is=True)

    @edge_case_handler
    async def setConfig(self,
        cacheSize=None,
        enableWebsocket=None
    ):
        return assertOperatorSuccess(
            await fetch.http_get(f"{self.baseurl}/config", {
                "sessionKey": self.session_key,
                **({
                    "cacheSize": cacheSize
                } if cacheSize else {}),
                **({
                    "enableWebsocket": enableWebsocket
                } if enableWebsocket else {})
            }
        ), raise_exception=True, return_as_is=True)

    @protocol_log
    @edge_case_handler
    async def sendFriendMessage(self,
        friend: T.Union[Friend, int],
        message: T.Union[
            MessageChain,
            BaseMessageComponent,
            T.List[T.Union[BaseMessageComponent, InternalImage]],
            str
        ]
    ) -> BotMessage:
        return BotMessage.parse_obj(assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/sendFriendMessage", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsFriend(friend),
                "messageChain": await self.handleMessageAsFriend(message)
            }
        ), raise_exception=True, return_as_is=True))
    
    @protocol_log
    @edge_case_handler
    async def sendGroupMessage(self,
        group: T.Union[Group, int],
        message: T.Union[
            MessageChain,
            BaseMessageComponent,
            T.List[T.Union[BaseMessageComponent, InternalImage]],
            str
        ],
        quoteSource: T.Union[int, components.Source]=None
    ) -> BotMessage:
        return BotMessage.parse_obj(assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/sendGroupMessage", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsGroup(group),
                "messageChain": await self.handleMessageAsGroup(message),
                **({"quote": quoteSource.id \
                    if isinstance(quoteSource, components.Source) else quoteSource}\
                if quoteSource else {})
            }
        ), raise_exception=True, return_as_is=True))

    @protocol_log
    @edge_case_handler
    async def revokeMessage(self, source: T.Union[components.Source, BotMessage, int]):
        return assertOperatorSuccess(await fetch.http_post(f"{self.baseurl}/recall", {
            "sessionKey": self.session_key,
            "target": source if isinstance(source, int) else source.id \
                if isinstance(source, components.Source) else source.messageId\
                if isinstance(source, BotMessage) else\
                    raiser(TypeError("invaild message source"))
        }), raise_exception=True)

    @protocol_log
    @edge_case_handler
    async def groupList(self) -> T.List[Group]:
        return [Group.parse_obj(group_info) \
            for group_info in await fetch.http_get(f"{self.baseurl}/groupList", {
                "sessionKey": self.session_key
            })
        ]

    @protocol_log
    @edge_case_handler
    async def friendList(self) -> T.List[Friend]:
        return [Friend.parse_obj(friend_info) \
            for friend_info in await fetch.http_get(f"{self.baseurl}/friendList", {
                "sessionKey": self.session_key
            })
        ]

    @protocol_log
    @edge_case_handler
    async def memberList(self, target: int) -> T.List[Member]:
        return [Member.parse_obj(member_info) \
            for member_info in await fetch.http_get(f"{self.baseurl}/memberList", {
                "sessionKey": self.session_key,
                "target": target
            })
        ]

    @protocol_log
    @edge_case_handler
    async def groupMemberNumber(self, target: int) -> int:
        return len(await self.memberList(target)) + 1

    @protocol_log
    @edge_case_handler
    async def uploadImage(self, type: T.Union[str, ImageType], image: InternalImage):
        post_result = json.loads(await fetch.upload(f"{self.baseurl}/uploadImage", image.render(), {
            "sessionKey": self.session_key,
            "type": type if isinstance(type, str) else type.value
        }))
        return components.Image(**post_result)

    @edge_case_handler
    async def fetchMessage(self, count: int) -> T.List[T.Union[FriendMessage, GroupMessage, ExternalEvent]]:
        from mirai.event.external.enums import ExternalEvents
        result = assertOperatorSuccess(
            await fetch.http_get(f"{self.baseurl}/fetchMessage", {
                "sessionKey": self.session_key,
                "count": count
            }
        ), raise_exception=True, return_as_is=True)
        # 因为重新生成一个开销太大, 所以就直接在原数据内进行遍历替换
        try:
            for index in range(len(result)):
                # 判断当前项是否为 Message
                if result[index]['type'] in MessageTypes:
                    if 'messageChain' in result[index]: 
                        result[index]['messageChain'] = MessageChain.parse_obj(result[index]['messageChain'])

                    result[index] = \
                        MessageTypes[result[index]['type']].parse_obj(result[index])

                elif hasattr(ExternalEvents, result[index]['type']):
                    # 判断当前项为 Event
                    result[index] = \
                        ExternalEvents[result[index]['type']].value.parse_obj(result[index])
        except pydantic.ValidationError:
            ProtocolLogger.error(f"parse failed: {result}")
            traceback.print_exc()
            raise
        return result

    @protocol_log
    @edge_case_handler
    async def messageFromId(self, sourceId: T.Union[components.Source, components.Quote, int]):
        if isinstance(sourceId, (components.Source, components.Quote)):
            sourceId = sourceId.id

        result = assertOperatorSuccess(await fetch.http_get(f"{self.baseurl}/messageFromId", {
            "sessionKey": self.session_key,
            "id": sourceId
        }), raise_exception=True, return_as_is=True)

        if result['type'] in MessageTypes:
            if "messageChain" in result:
                result['messageChain'] = MessageChain.custom_parse(result['messageChain'])

            return MessageTypes[result['type']].parse_obj(result)
        else:
            raise TypeError(f"unknown message, not found type.")

    @protocol_log
    @edge_case_handler
    async def muteAll(self, group: T.Union[Group, int]) -> bool:
        return assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/muteAll", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsGroup(group)
            }
        ), raise_exception=True)

    @protocol_log
    @edge_case_handler
    async def unmuteAll(self, group: T.Union[Group, int]) -> bool:
        return assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/unmuteAll", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsGroup(group)
            }
        ), raise_exception=True)
    
    @protocol_log
    @edge_case_handler
    async def memberInfo(self,
        group: T.Union[Group, int],
        member: T.Union[Member, int]
    ):
        return MemberChangeableSetting.parse_obj(assertOperatorSuccess(
            await fetch.http_get(f"{self.baseurl}/memberInfo", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsGroup(group),
                "memberId": self.handleTargetAsMember(member)
            }
        ), raise_exception=True, return_as_is=True))

    @protocol_log
    @edge_case_handler
    async def botMemberInfo(self,
        group: T.Union[Group, int]
    ):
        return await self.memberInfo(group, self.qq)

    @protocol_log
    @edge_case_handler
    async def changeMemberInfo(self,
        group: T.Union[Group, int],
        member: T.Union[Member, int],
        setting: MemberChangeableSetting
    ) -> bool:
        return assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/memberInfo", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsGroup(group),
                "memberId": self.handleTargetAsMember(member),
                "info": json.loads(setting.json())
            }
        ), raise_exception=True)

    @protocol_log
    @edge_case_handler
    async def groupConfig(self, group: T.Union[Group, int]) -> GroupSetting:
        return GroupSetting.parse_obj(
            await fetch.http_get(f"{self.baseurl}/groupConfig", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsGroup(group)
            })
        )

    @protocol_log
    @edge_case_handler
    async def changeGroupConfig(self,
        group: T.Union[Group, int],
        config: GroupSetting
    ) -> bool:
        return assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/groupConfig", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsGroup(group),
                "config": json.loads(config.json())
            }
        ), raise_exception=True)

    @protocol_log
    @edge_case_handler
    async def mute(self,
        group: T.Union[Group, int],
        member: T.Union[Member, int],
        time: T.Union[timedelta, int]
    ):
        if isinstance(time, timedelta):
            time = int(time.total_seconds())
        time = min(86400 * 30, max(60, time))
        return assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/mute", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsGroup(group),
                "memberId": self.handleTargetAsMember(member),
                "time": time
            }
        ), raise_exception=True)

    @protocol_log
    @edge_case_handler
    async def unmute(self,
        group: T.Union[Group, int],
        member: T.Union[Member, int]
    ):
        return assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/unmute", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsGroup(group),
                "memberId": self.handleTargetAsMember(member),
            }
        ), raise_exception=True)

    @protocol_log
    @edge_case_handler
    async def kick(self,
        group: T.Union[Group, int],
        member: T.Union[Member, int],
        kickMessage: T.Optional[str] = None
    ):
        return assertOperatorSuccess(
            await fetch.http_post(f"{self.baseurl}/kick", {
                "sessionKey": self.session_key,
                "target": self.handleTargetAsGroup(group),
                "memberId": self.handleTargetAsMember(member),
                **({
                    "msg": kickMessage
                } if kickMessage else {})
            }
        ), raise_exception=True)

    async def handleMessageAsGroup(
        self,
        message: T.Union[
            MessageChain,
            BaseMessageComponent,
            T.List[T.Union[BaseMessageComponent, InternalImage]],
            str
        ]):
        if isinstance(message, MessageChain):
            return json.loads(message.json())
        elif isinstance(message, BaseMessageComponent):
            return [json.loads(message.json())]
        elif isinstance(message, (tuple, list)):
            result = []
            for i in message:
                if isinstance(i, InternalImage):
                    result.append({
                        "type": "Image",
                        "imageId": (await self.handleInternalImageAsGroup(i)).asGroupImage()
                    })
                elif isinstance(i, components.Image):
                    result.append({
                        "type": "Image",
                        "imageId": i.asGroupImage()
                    })
                else:
                    result.append(json.loads(i.json()))
            return result
        elif isinstance(message, str):
            return [json.loads(components.Plain(text=message).json())]
        else:
            raise raiser(ValueError("invaild message."))

    async def handleMessageAsFriend(
        self,
        message: T.Union[
            MessageChain,
            BaseMessageComponent,
            T.List[BaseMessageComponent],
            str
        ]):
        if isinstance(message, MessageChain):
            return json.loads(message.json())
        elif isinstance(message, BaseMessageComponent):
            return [json.loads(message.json())]
        elif isinstance(message, (tuple, list)):
            result = []
            for i in message:
                if isinstance(i, InternalImage):
                    result.append({
                        "type": "Image",
                        "imageId": (await self.handleInternalImageAsGroup(i)).asFriendImage()
                    })
                elif isinstance(i, components.Image):
                    result.append({
                        "type": "Image",
                        "imageId": i.asGroupImage()
                    })
                else:
                    result.append(json.loads(i.json()))
            return result
        elif isinstance(message, str):
            return [json.loads(components.Plain(text=message).json())]
        else:
            raise raiser(ValueError("invaild message."))

    def handleTargetAsGroup(self, target: T.Union[Group, int]):
        return target if isinstance(target, int) else \
            target.id if isinstance(target, Group) else \
                raiser(ValueError("invaild target as group."))

    def handleTargetAsFriend(self, target: T.Union[Friend, int]):
        return target if isinstance(target, int) else \
            target.id if isinstance(target, Friend) else \
                raiser(ValueError("invaild target as a friend obj."))

    def handleTargetAsMember(self, target: T.Union[Member, int]):
        return target if isinstance(target, int) else \
            target.id if isinstance(target, Member) else \
                raiser(ValueError("invaild target as a member obj."))

    async def handleInternalImageAsGroup(self, image: InternalImage):
        return await self.uploadImage("group", image)

    async def handleInternalImageAsFriend(self, image: InternalImage):
        return await self.uploadImage("friend", image)


from logbook import Logger, StreamHandler
from logbook import (
    INFO,
    DEBUG
)
import os
import sys

stream_handler = StreamHandler(sys.stdout, level=INFO if not os.environ.get("MIRAI_DEBUG") else DEBUG)
stream_handler.format_string = '[{record.time:%Y-%m-%d %H:%M:%S}][Mirai] {record.level_name}: {record.channel}: {record.message}'
stream_handler.push_application()

Event = Logger('Event', level=INFO)
Network = Logger("Network", level=DEBUG)
Session = Logger("Session", level=INFO)
Protocol = Logger("Protocol", level=INFO)


import json
import mimetypes
import typing as T
from pathlib import Path
from .logger import Network

import aiohttp

from mirai.exceptions import NetworkError

class fetch:
    @staticmethod
    async def http_post(url, data_map):
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=data_map) as response:
                response.raise_for_status()
                data = await response.text(encoding="utf-8")
                Network.debug(f"requested url={url}, by data_map={data_map}, and status={response.status}, data={data}")
        try:
            return json.loads(data)
        except json.decoder.JSONDecodeError:
            Network.error(f"requested {url} with {data_map}, responsed {data}, decode failed...")

    @staticmethod
    async def http_get(url, params=None): 
        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=params) as response:
                response.raise_for_status()
                data = await response.text(encoding="utf-8")
                Network.debug(f"requested url={url}, by params={params}, and status={response.status}, data={data}")
        try:
            return json.loads(data)
        except json.decoder.JSONDecodeError:
            Network.error(f"requested {url} with {params}, responsed {data}, decode failed...")

    @staticmethod
    async def upload(url, filedata: bytes, addon_dict: dict):
        upload_data = aiohttp.FormData()
        upload_data.add_field("img", filedata)
        for item in addon_dict.items():
            upload_data.add_fields(item)

        async with aiohttp.ClientSession() as session:
            async with session.post(url, data=upload_data) as response:
                response.raise_for_status()
                Network.debug(f"requested url={url}, and status={response.status}, addon_dict={addon_dict}")
                return await response.text("utf-8")


from pathlib import Path
from abc import ABCMeta, abstractmethod
import base64

class InternalImage(metaclass=ABCMeta):
    @abstractmethod
    def __init__(self):
        super().__init__()

    @abstractmethod
    def render(self) -> bytes:
        pass

class LocalImage(InternalImage):
    path: Path

    def __init__(self, path):
        if isinstance(path, str):
            self.path = Path(path)
        elif isinstance(path, Path):
            self.path = path

    def render(self) -> bytes:
        return self.path.read_bytes()

class IOImage(InternalImage):
    def __init__(self, IO):
        """make a object with 'read' method a image.

        IO - a object, must has a `read` method to return bytes.
        """
        self.IO = IO

    def render(self) -> bytes:
        return self.IO.read()

class BytesImage(InternalImage):
    def __init__(self, data: bytes):
        self.data = data

    def render(self) -> bytes:
        return self.data

class Base64Image(InternalImage):
    def __init__(self, base64_str):
        self.base64_str = base64_str
    
    def render(self) -> bytes:
        return base64.b64decode(self.base64_str)


import typing as T
from enum import Enum
from .base import MessageComponentTypes
from mirai.entities.friend import Friend
from mirai.entities.group import Group, Member
from pydantic import BaseModel
from .chain import MessageChain

class MessageItemType(Enum):
    FriendMessage = "FriendMessage"
    GroupMessage = "GroupMessage"
    BotMessage = "BotMessage"

class FriendMessage(BaseModel):
    type: MessageItemType = "FriendMessage"
    messageChain: T.Optional[MessageChain]
    sender: Friend

    def toString(self):
        if self.messageChain:
            return self.messageChain.toString()

class GroupMessage(BaseModel):
    type: MessageItemType = "GroupMessage"
    messageChain: T.Optional[MessageChain]
    sender: Member

    def toString(self):
        if self.messageChain:
            return self.messageChain.toString()

class BotMessage(BaseModel):
    type: MessageItemType = 'BotMessage'
    messageId: int

MessageTypes = {
    "GroupMessage": GroupMessage,
    "FriendMessage": FriendMessage
}


from mirai.application import Mirai
from mirai.event.message.chain import MessageChain
from mirai import (
    Member, Friend,
    Group, Any
)
from typing import (
    List, Dict, Callable
)
from .entity import Command
import parse

class CommandManager:
    main_application: Mirai
    command_prefix: str = "/"
    matches_commands: List[Command] = {}

    def __init__(self,
        app: Mirai,

        group_message: bool = True,
        friend_message: bool = True,

        command_prefix: str = "/"
    ):
        self.main_application = app
        self.command_prefix = command_prefix
        if group_message:
            app.receiver("GroupMessage")(self.group_message_handler)
        if friend_message:
            app.receiver("FriendMessage")(self.friend_message_handler)

    def sortCommands(self):
        self.matches_commands.sort(key=lambda x: x.priority, reverse=True)

    async def group_message_handler(self,
        app: Mirai, message: MessageChain,
        sender: Member, group: Group
    ):
        message_string = message.toString()
        for command in self.matches_commands:
            pass

    async def friend_message_handler(self):
        pass


from enum import Enum
from dataclasses import dataclass
from pydantic import BaseModel

class Permission(Enum):
    Member = "MEMBER"
    Administrator = "ADMINISTRATOR"
    Owner = "OWNER"

class Group(BaseModel):
    id: int
    name: str
    permission: Permission

    def __repr__(self):
        return f"<Group id={self.id} name='{self.name}' permission={self.permission.name}>"

    def getAvatarUrl(self) -> str:
        return f'https://p.qlogo.cn/gh/{self.id}/{self.id}_1/140'

class Member(BaseModel):
    id: int
    memberName: str
    permission: Permission
    group: Group

    def __repr__(self):
        return f"<GroupMember id={self.id} group={self.group} permission={self.permission} group={self.group.id}>"

    def getAvatarUrl(self) -> str:
        return f'http://q4.qlogo.cn/g?b=qq&nk={self.id}&s=140'

class MemberChangeableSetting(BaseModel):
    name: str
    specialTitle: str

    def modify(self, **kwargs):
        for i in ("name", "kwargs"):
            if i in kwargs:
                setattr(self, i, kwargs[i])
        return self

class GroupSetting(BaseModel):
    name: str
    announcement: str
    confessTalk: bool
    allowMemberInvite: bool
    autoApprove: bool
    anonymousChat: bool

    def modify(self, **kwargs):
        for i in ("name",
            "announcement",
            "confessTalk",
            "allowMemberInvite",
            "autoApprove",
            "anonymousChat"
        ):
            if i in kwargs:
                setattr(self, i, kwargs[i])
        return self


import typing as T
from mirai.misc import findKey, printer, ImageRegex, getMatchedString
from mirai.event.message.base import BaseMessageComponent, MessageComponentTypes
from pydantic import Field, validator, HttpUrl
from io import BytesIO
from pathlib import Path
from mirai.image import (
    LocalImage,
    IOImage,
    Base64Image, BytesImage,
)
from mirai.logger import Protocol as ProtocolLogger
from aiohttp import ClientSession
import datetime

__all__ = [
    "Plain",
    "Source",
    "At",
    "AtAll",
    "Face",
    "Image",
    "Unknown",
    "Quote"
]

class Plain(BaseMessageComponent):
    type: MessageComponentTypes = "Plain"
    text: str

    def __init__(self, text, **_):
        if len(text) > 128:
            ProtocolLogger.warn(f"mirai does not support for long string: now its length is {len(text)}")
        super().__init__(text=text, type="Plain")

    def toString(self):
        return self.text

class Source(BaseMessageComponent):
    type: MessageComponentTypes = "Source"
    id: int
    time: datetime.datetime

    def toString(self):
        return ""

from .chain import MessageChain

class Quote(BaseMessageComponent):
    type: MessageComponentTypes = "Quote"
    id: T.Optional[int]
    groupId: T.Optional[int]
    senderId: T.Optional[int]
    origin: MessageChain

    @validator("origin", always=True, pre=True)
    @classmethod
    def origin_formater(cls, v):
        return MessageChain.parse_obj(v)

    def __init__(self, id: int, groupId: int, senderId: int, origin: int, **_):
        super().__init__(
            id=id,
            groupId=groupId,
            senderId=senderId,
            origin=origin
        )

    def toString(self):
        return ""

class At(BaseMessageComponent):
    type: MessageComponentTypes = "At"
    target: int
    display: T.Optional[str] = None

    def __init__(self, target, display=None, **_):
        super().__init__(target=target, display=display)

    def toString(self):
        return f"[At::target={self.target}]"

class AtAll(BaseMessageComponent):
    type: MessageComponentTypes = "AtAll"

    def __init__(self, **_):
        super().__init__()

    def toString(self):
        return f"[AtAll]"

class Face(BaseMessageComponent):
    type: MessageComponentTypes = "Face"
    faceId: int
    name: T.Optional[str]

    def __init__(self, faceId, name=None, **_):
        super().__init__(faceId=faceId, name=name)

    def toString(self):
        return f"[Face::name={self.name}]"

class Image(BaseMessageComponent):
    type: MessageComponentTypes = "Image"
    imageId: T.Optional[str]
    url: T.Optional[HttpUrl] = None

    @validator("imageId", always=True, pre=True)
    @classmethod
    def imageId_formater(cls, v):
        length = len(v)
        if length == 42:
            # group
            return v[1:-5]
        elif length == 37:
            return v[1:]
        else:
            return v

    def __init__(self, imageId, url=None, **_):
        super().__init__(imageId=imageId, url=url)

    def toString(self):
        return f"[Image::{self.imageId}]"

    def asGroupImage(self) -> str:
        return f"{{{self.imageId.upper()}}}.jpg"

    def asFriendImage(self) -> str:
        return f"/{self.imageId.lower()}"

    @staticmethod
    def fromFileSystem(path: T.Union[Path, str]) -> LocalImage:
        return LocalImage(path)

    async def toBytes(self, chunk_size=256) -> BytesIO:
        async with ClientSession() as session:
            async with session.get(self.url) as response:
                result = BytesIO()
                while True:
                    chunk = await response.content.read(chunk_size)
                    if not chunk:
                        break
                    result.write(chunk)
            return result

    @staticmethod
    def fromBytes(data) -> BytesImage:
        return BytesImage(data)

    @staticmethod
    def fromBase64(base64_str) -> Base64Image:
        return Base64Image(base64_str)

    @staticmethod
    def fromIO(IO) -> IOImage:
        return IOImage(IO)

class Xml(BaseMessageComponent):
    type: MessageComponentTypes = "Xml"
    XML: str

    def __init__(self, xml, type="Xml"):
        super().__init__(XML=xml)

class Json(BaseMessageComponent):
    type: MessageComponentTypes = "Json"
    Json: dict = Field(..., alias="json")

    def __init__(self, json: dict, **_):
        super().__init__(Json=json)

class App(BaseMessageComponent):
    type: MessageComponentTypes = "App"
    content: str

    def __init__(self, content: str, **_):
        super().__init__(content=content)

class Unknown(BaseMessageComponent):
    type: MessageComponentTypes = "Unknown"
    text: str

    def toString(self):
        return ""

MessageComponents = {
    "At": At,
    "AtAll": AtAll,
    "Face": Face,
    "Plain": Plain,
    "Image": Image,
    "Source": Source,
    "Quote": Quote,
    "Xml": Xml,
    "Json": Json,
    "App": App,
    "Unknown": Unknown
}


from enum import Enum

class ExternalEventTypes(Enum):
    BotOnlineEvent = "BotOnlineEvent"
    BotOfflineEventActive = "BotOfflineEventActive"
    BotOfflineEventForce = "BotOfflineEventForce"
    BotOfflineEventDropped = "BotOfflineEventDropped"
    BotReloginEvent = "BotReloginEvent"
    BotGroupPermissionChangeEvent = "BotGroupPermissionChangeEvent"
    BotMuteEvent = "BotMuteEvent"
    BotUnmuteEvent = "BotUnmuteEvent"
    BotJoinGroupEvent = "BotJoinGroupEvent"

    GroupRecallEvent = "GroupRecallEvent"
    FriendRecallEvent = "FriendRecallEvent"

    GroupNameChangeEvent = "GroupNameChangeEvent"
    GroupEntranceAnnouncementChangeEvent = "GroupEntranceAnnouncementChangeEvent"
    GroupMuteAllEvent = "GroupMuteAllEvent"

    # 群设置被修改事件
    GroupAllowAnonymousChatEvent = "GroupAllowAnonymousChatEvent" # 群设置 是否允许匿名聊天 被修改
    GroupAllowConfessTalkEvent = "GroupAllowConfessTalkEvent" # 坦白说
    GroupAllowMemberInviteEvent = "GroupAllowMemberInviteEvent" # 邀请进群

    # 群事件(被 Bot 监听到的, 为"被动事件", 其中 Bot 身份为第三方.)
    MemberJoinEvent = "MemberJoinEvent"
    MemberLeaveEventKick = "MemberLeaveEventKick"
    MemberLeaveEventQuit = "MemberLeaveEventQuit"
    MemberCardChangeEvent = "MemberCardChangeEvent"
    MemberSpecialTitleChangeEvent = "MemberSpecialTitleChangeEvent"
    MemberPermissionChangeEvent = "MemberPermissionChangeEvent"
    MemberMuteEvent = "MemberMuteEvent"
    MemberUnmuteEvent = "MemberUnmuteEvent"

    # python-mirai 自己提供的事件
    UnexceptedException = "UnexceptedException"


from enum import Enum
import asyncio
from threading import Thread, Lock
import typing as T
import random
from .logger import Protocol, Session
import os
import re
import aiohttp
from . import exceptions
import traceback

def assertOperatorSuccess(result, raise_exception=False, return_as_is=False):
  if not result:
    if raise_exception:
      raise exceptions.InvaildSession("this method returned None, as sessionkey invaild...")
    else:
      return None
  if "code" in result:
    if not raise_exception:
      return result['code'] == 0
    else:
      if result['code'] != 0:
        raise {
          1: exceptions.AuthenticateError, # 这种情况需要检查Authkey, 可能还是连错了.
          2: exceptions.LoginException, # 嗯...你是不是忘记在mirai-console登录了?...算了 自动重连.
          3: exceptions.InvaildSession, # 这种情况会自动重连.
          4: exceptions.ValidatedSession, # 啊 smjb错误... 也会自动重连
          5: exceptions.UnknownReceiverTarget, # 业务代码错误.
          10: PermissionError, # 一般业务代码错误, 自行亦会
          400: exceptions.CallDevelopers # 发生这个错误...你就给我提个ISSUE
        }[result['code']](f"""invaild stdin: { {
          1: "wrong auth key",
          2: "unknown qq account",
          3: "invaild session key",
          4: "disabled session key",
          5: "unknown receiver target",
          10: "permission denied",
          400: "wrong arguments"
        }[result['code']] }""")
      else:
        if return_as_is:
          return result
        else:
          return True
  if return_as_is:
    return result
  return False

class ImageType(Enum):
  Friend = "friend"
  Group = "group"

TRACEBACKED = os.urandom(32)

ImageRegex = {
  "group": r"(?<=\{)([0-9A-Z]{8})\-([0-9A-Z]{4})-([0-9A-Z]{4})-([0-9A-Z]{4})-([0-9A-Z]{12})(?=\}\..*?)",
  "friend": r"(?<=/)([0-9a-z]{8})\-([0-9a-z]{4})-([0-9a-z]{4})-([0-9a-z]{4})-([0-9a-z]{12})"
}

_windows_device_files = (
  "CON",
  "AUX",
  "COM1",
  "COM2",
  "COM3",
  "COM4",
  "LPT1",
  "LPT2",
  "LPT3",
  "PRN",
  "NUL",
)
_filename_ascii_strip_re = re.compile(r"[^A-Za-z0-9_.-]")

def getMatchedString(regex_result):
  if regex_result:
    return regex_result.string[slice(*regex_result.span())]

def findKey(mapping, value):
  try:
    index = list(mapping.values()).index(value)
  except ValueError:
    return "Unknown"
  return list(mapping.keys())[index]

def raiser(error):
  raise error

def printer(val):
  print(val)
  return val

def justdo(call, val):
  print(call())
  return val

def randomNumberString():
  return str(random.choice(range(100000000, 9999999999)))

def randomRangedNumberString(length_range=(9,)):
  length = random.choice(length_range)
  return random.choice(range(10**(length - 1), int("9"*(length))))

def protocol_log(func):
  async def wrapper(*args, **kwargs):
    try:
      result = await func(*args, **kwargs)
      Protocol.info(f"protocol method {func.__name__} was called")
      return result
    except Exception as e:
      Protocol.error(f"protocol method {func.__name__} raised a error: {e.__class__.__name__}")
      raise e
  return wrapper

def secure_filename(filename):
  if isinstance(filename, str):
    from unicodedata import normalize

    filename = normalize("NFKD", filename).encode("ascii", "ignore")
    filename = filename.decode("ascii")

  for sep in os.path.sep, os.path.altsep:
    if sep:
      filename = filename.replace(sep, " ")

  filename = \
    str(_filename_ascii_strip_re.sub("", "_".join(filename.split()))).strip("._")

  if (
    os.name == "nt" and filename and \
    filename.split(".")[0].upper() in _windows_device_files
  ):
    filename = "_" + filename

  return filename

def edge_case_handler(func):
  async def wrapper(self, *args, **kwargs):
    while True:
      try:
        return await func(self, *args, **kwargs)
      except exceptions.AuthenticateError:
        Protocol.error("invaild authkey, please check your input.")
        exit(1)
      except exceptions.LoginException:
        Protocol.error("there is not such qq in headless client, we will try again after 5 seconds.")
        await asyncio.sleep(5)
        if func.__name__ != "verify":
          await self.verify()
        continue
      except (exceptions.InvaildSession, exceptions.ValidatedSession):
        Protocol.error("a unexpected session error, we will deal with it.")
        await self.enable_session()
      except aiohttp.client_exceptions.ClientError:
        Protocol.error(f"cannot connect to the headless client, will retry after 5 seconds.")
        await asyncio.sleep(5)
        continue
      except exceptions.CallDevelopers:
        Protocol.error("emmm, please contect me at github.")
        exit(-1)
      except:
        raise
  wrapper.__name__ = func.__name__
  return wrapper

def if_error_print_arg(func):
  def wrapper(*args, **kwargs):
    try:
      return func(*args, **kwargs)
    except:
      print(args, kwargs)
      traceback.print_exc()
  return wrapper


import typing as T
from pydantic import BaseModel

from .base import BaseMessageComponent
from mirai.misc import raiser, printer, if_error_print_arg
from .components import Source
from mirai.logger import Protocol

class MessageChain(BaseModel):
    __root__: T.List[BaseMessageComponent] = []

    def __add__(self, value):
        if isinstance(value, BaseMessageComponent):
            self.__root__.append(value)
            return self
        elif isinstance(value, MessageChain):
            self.__root__ += value.__root__
            return self

    def toString(self) -> str:
        return "".join([i.toString() for i in self.__root__])

    @classmethod
    def parse_obj(cls, obj):
        from .components import MessageComponents
        result = []
        for i in obj:
            if not isinstance(i, dict):
                raise TypeError("invaild value")
            try:
                result.append(MessageComponents[i['type']].parse_obj(i))
            except:
                Protocol.error(f"error throwed by message serialization: {i['type']}, it's {i}")
                raise
        return cls(__root__=result)

    def __iter__(self):
        yield from self.__root__

    def __getitem__(self, index):
        return self.__root__[index]

    def hasComponent(self, component_class) -> bool:
        for i in self:
            if type(i) == component_class:
                return True
        else:
            return False

    def __len__(self) -> int:
        return len(self.__root__)

    def getFirstComponent(self, component_class) -> T.Optional[BaseMessageComponent]:
        for i in self:
            if type(i) == component_class:
                return i

    def getAllofComponent(self, component_class) -> T.List[BaseMessageComponent]:
        return [i for i in self if type(i) == component_class]

    def getSource(self) -> Source:
        return self.getFirstComponent(Source)


class NetworkError(Exception):
    pass

class Cancelled(Exception):
    pass

class UnknownTarget(Exception):
    pass

class LoginException(Exception):
    "你忘记在mirai-console登录就这种错误."
    pass

class AuthenticateError(Exception):
    pass

class InvaildSession(Exception):
    pass

class ValidatedSession(Exception):
    "一般来讲 这种情况出现时需要刷新session"

class UnknownReceiverTarget(Exception):
    pass

class CallDevelopers(Exception):
    '还愣着干啥?开ISSUE啊!'
    pass


from collections import namedtuple
from typing import Any
from enum import Enum
from pydantic import BaseModel

# 内部事件实现.
InternalEvent = namedtuple("Event", ("name", "body"))

from .enums import ExternalEventTypes
class ExternalEvent(BaseModel):
    type: ExternalEventTypes


import asyncio
import copy
import inspect
import traceback
from functools import partial
from typing import (
    Any, Awaitable, Callable, Dict, List, NamedTuple, Optional, Union)
from urllib import parse
import contextlib

import pydantic
import aiohttp

from mirai.depend import Depend
from mirai.entities.friend import Friend
from mirai.entities.group import Group, Member
from mirai.event import ExternalEvent, ExternalEventTypes, InternalEvent
from mirai.event.message import MessageChain, components
from mirai.event.message.models import (FriendMessage, GroupMessage,
                                        MessageItemType, MessageTypes)
from mirai.logger import Event as EventLogger
from mirai.logger import Session as SessionLogger
from mirai.misc import raiser, TRACEBACKED
from mirai.network import fetch
from mirai.protocol import MiraiProtocol
from mirai.exceptions import Cancelled

class Mirai(MiraiProtocol):
  event: Dict[
    str, List[Callable[[Any], Awaitable]]
  ] = {}
  subroutines: List[Callable] = []
  lifecycle: Dict[str, List[Callable]] = {
    "start": [],
    "end": [],
    "around": []
  }
  useWebsocket = False
  listening_exceptions: List[Exception] = []
  extensite_config: Dict
  global_dependencies: List[Depend]

  def __init__(self,
    url: Optional[str] = None,

    host: Optional[str] = None,
    port: Optional[int] = None,
    authKey: Optional[str] = None,
    qq: Optional[int] = None,

    websocket: bool = False,
    extensite_config: dict = None,
    global_dependencies: List[Depend] = None
  ):
    self.extensite_config = extensite_config or {}
    self.global_dependencies = global_dependencies or []
    self.useWebsocket = websocket
    if url:
      urlinfo = parse.urlparse(url)
      if urlinfo:
        query_info = parse.parse_qs(urlinfo.query)
        if all([
          urlinfo.scheme == "mirai",
          urlinfo.path in ["/", "/ws"],

          "authKey" in query_info and query_info["authKey"],
          "qq" in query_info and query_info["qq"]
        ]):
          if urlinfo.path == "/ws":
            self.useWebsocket = True
          else:
            self.useWebsocket = websocket

          authKey = query_info["authKey"][0]

          self.baseurl = f"http://{urlinfo.netloc}"
          self.auth_key = authKey
          self.qq = query_info["qq"][0]
        else:
          raise ValueError("invaild url: wrong format")
      else:
        raise ValueError("invaild url")
    else:
      if all([host, port, authKey, qq]):
        self.baseurl = f"http://{host}:{port}"
        self.auth_key = authKey
        self.qq = qq
      else:
        raise ValueError("invaild arguments")

  async def enable_session(self) -> "Session":
    auth_response = await self.auth()
    if all([
      "code" in auth_response and auth_response['code'] == 0,
      "session" in auth_response and auth_response['session']
    ]):
      if "msg" in auth_response and auth_response['msg']:
        self.session_key = auth_response['msg']
      else:
        self.session_key = auth_response['session']

      await self.verify()
    else:
      if "code" in auth_response and auth_response['code'] == 1:
        raise ValueError("invaild authKey")
      else:
        raise ValueError('invaild args: unknown response')

    self.enabled = True
    return self

  def receiver(self, event_name,
      dependencies: List[Depend] = None,
      use_middlewares: List[Callable] = None
    ):
    def receiver_warpper(
      func: Callable[[Union[FriendMessage, GroupMessage], "Session"], Awaitable[Any]]
    ):
      if not inspect.iscoroutinefunction(func):
        raise TypeError("event body must be a coroutine function.")

      protocol = {
        "func": func,
        "dependencies": dependencies or [],
        "middlewares": use_middlewares or []
      }

      protocol['dependencies'] += self.global_dependencies
      # support for global dependencies
      
      if event_name not in self.event:
        self.event[event_name] = [protocol]
      else:
        self.event[event_name].append(protocol)
      return func
    return receiver_warpper

  async def throw_exception_event(self, event_context, queue, exception):
    from .event.builtins import UnexpectedException
    if event_context.name != "UnexpectedException":
      #print("error: by pre:", event_context.name)
      await queue.put(InternalEvent(
        name="UnexpectedException",
        body=UnexpectedException(
          error=exception,
          event=event_context,
          application=self
        )
      ))
      EventLogger.error(f"threw a exception by {event_context.name}, Exception: {exception.__class__.__name__}, but it has been catched.")
    else:
      EventLogger.critical(f"threw a exception by {event_context.name}, Exception: {exception.__class__.__name__}, it's a exception handler.")

  async def argument_compiler(self, func, event_context):
    annotations_mapping = self.get_annotations_mapping()
    signature_mapping = self.signature_getter(func)
    translated_mapping = { # 执行主体
      k: annotations_mapping[v](
        event_context
      )\
      for k, v in func.__annotations__.items()\
        if \
          k != "return" and \
          k not in signature_mapping # 嗯...你设了什么default? 放你过去.
    }
    return translated_mapping

  @staticmethod
  def signature_getter(func):
    "获取函数的默认值列表"
    return {k: v.default \
      for k, v in dict(inspect.signature(func).parameters).items() \
        if v.default != inspect._empty}

  @staticmethod
  def signature_checker(func):
    signature_mapping = Mirai.signature_getter(func)
    for i in signature_mapping.values():
      if not isinstance(i, Depend):
        raise TypeError("you must use a Depend to patch the default value.")
    return signature_mapping

  async def signature_checkout(self, func, event_context, queue):
    signature_mapping = self.signature_getter(func)
    return {
      k: await self.main_entrance(
        v.func,
        event_context,
        queue
      ) for k, v in signature_mapping.items()
    }

  async def main_entrance(self, run_body, event_context, queue):
    if isinstance(run_body, dict):
      callable_target = run_body['func']
      for depend in run_body['dependencies']:
        if not inspect.isclass(depend.func):
          depend_func = depend.func
        elif hasattr(depend.func, "__call__"):
          depend_func = depend.func.__call__
        else:
          raise TypeError("must be callable.")
        
        try:
          result = await self.main_entrance(
            {
              "func": depend_func,
              "middlewares": depend.middlewares,
              "dependencies": []
            },
            event_context, queue
          )
          if result is TRACEBACKED:
            return TRACEBACKED
        except Cancelled:
          return TRACEBACKED
        except (NameError, TypeError) as e:
          EventLogger.error(f"threw a exception by {event_context.name}, it's about Annotations Checker, please report to developer.")
          traceback.print_exc()
        except Exception as e:
          if type(e) not in self.listening_exceptions:
            EventLogger.error(f"threw a exception by {event_context.name} in a depend, and it's {e}, body has been cancelled.")
            raise
          else:
            await self.throw_exception_event(event_context, queue, e)
            return TRACEBACKED

    else:
      if inspect.isclass(run_body):
        if hasattr(run_body, "__call__"):
          run_body = run_body.__call__
        else:
          raise TypeError("must be callable.")
      else:
        callable_target = run_body

    depend_handler_result = await self.signature_checkout(
      callable_target,
      event_context,
      queue
    )
    if depend_handler_result is TRACEBACKED:
      return TRACEBACKED

    translated_mapping = {
      **(await self.argument_compiler(
        callable_target,
        event_context
      )),
      **depend_handler_result
    }

    try:
      if isinstance(run_body, dict):
        middlewares = run_body.get("middlewares")
        if middlewares:
          async_middlewares = []
          normal_middlewares = []

          for middleware in middlewares:
            if all([
              hasattr(middleware, "__aenter__"),
              hasattr(middleware, "__aexit__")
            ]):
              async_middlewares.append(middleware)
            elif all([
              hasattr(middleware, "__enter__"),
              hasattr(middleware, "__exit__")
            ]):
              normal_middlewares.append(middleware)
            else:
              SessionLogger.error(f"threw a exception by {event_context.name}, no currect context error.")
              raise AttributeError("no a currect context object.")

          async with contextlib.AsyncExitStack() as async_stack:
            for async_middleware in async_middlewares:
              SessionLogger.debug(f"a event called {event_context.name}, enter a currect async context.")
              await async_stack.enter_async_context(async_middleware)

            with contextlib.ExitStack() as normal_stack:
              for normal_middleware in normal_middlewares:
                SessionLogger.debug(f"a event called {event_context.name}, enter a currect context.")
                normal_stack.enter_context(normal_middleware)

              if inspect.iscoroutinefunction(callable_target):
                return await callable_target(**translated_mapping)
              else:
                return callable_target(**translated_mapping)
        else:
          if inspect.iscoroutinefunction(callable_target):
            return await callable_target(**translated_mapping)
          else:
            return callable_target(**translated_mapping)
      else:
        if inspect.iscoroutinefunction(callable_target):
          return await callable_target(**translated_mapping)
        else:
          return callable_target(**translated_mapping)
    except (NameError, TypeError) as e:
      EventLogger.error(f"threw a exception by {event_context.name}, it's about Annotations Checker, please report to developer.")
      traceback.print_exc()
    except Cancelled:
      return TRACEBACKED
    except Exception as e:
      if type(e) not in self.listening_exceptions:
        EventLogger.error(f"threw a exception by {event_context.name}, and it's {e}")
        raise
      else:
        await self.throw_exception_event(event_context, queue, e)
        return TRACEBACKED

  async def message_polling(self, exit_signal, queue, count=10):
    while not exit_signal():
      await asyncio.sleep(0.5)

      try:
        result  = \
          await super().fetchMessage(count)
      except pydantic.ValidationError:
        continue
      last_length = len(result)
      latest_result = []
      while True:
        if last_length == count:
          latest_result = await super().fetchMessage(count)
          last_length = len(latest_result)
          result += latest_result
          continue
        break

      for message_index in range(len(result)):
        item = result[message_index]
        await queue.put(
          InternalEvent(
            name=self.getEventCurrentName(type(item)),
            body=item
          )
        )

  async def ws_event_receiver(self, exit_signal, queue):
    from mirai.event.external.enums import ExternalEvents
    async with aiohttp.ClientSession() as session:
      async with session.ws_connect(
        f"{self.baseurl}/all?sessionKey={self.session_key}"
      ) as ws_connection:
        while not exit_signal():
          try:
            received_data = await ws_connection.receive_json()
          except TypeError:
            if not exit_signal():
              continue
            else:
              break
          if received_data:
            try:
              if received_data['type'] in MessageTypes:
                  if 'messageChain' in received_data: 
                    received_data['messageChain'] = \
                      MessageChain.parse_obj(received_data['messageChain'])

                  received_data = \
                    MessageTypes[received_data['type']].parse_obj(received_data)

              elif hasattr(ExternalEvents, received_data['type']):
                  # 判断当前项是否为 Event
                  received_data = \
                    ExternalEvents[received_data['type']]\
                      .value\
                      .parse_obj(received_data)
            except pydantic.ValidationError:
              SessionLogger.error(f"parse failed: {received_data}")
              traceback.print_exc()
            else:
              await queue.put(InternalEvent(
                name=self.getEventCurrentName(type(received_data)),
                body=received_data
              ))

  async def event_runner(self, exit_signal_status, queue: asyncio.Queue):
    while not exit_signal_status():
      event_context: InternalEvent
      try:
        event_context: NamedTuple[InternalEvent] = await asyncio.wait_for(queue.get(), 3)
      except asyncio.TimeoutError:
        if exit_signal_status():
          break
        else:
          continue

      if event_context.name in self.registeredEventNames:
        for event_body in list(self.event.values())\
              [self.registeredEventNames.index(event_context.name)]:
          if event_body:
            EventLogger.info(f"handling a event: {event_context.name}")

            running_loop = asyncio.get_running_loop()
            running_loop.create_task(self.main_entrance(
              event_body,
              event_context, queue
            ))
  
  def getRestraintMapping(self):
    from mirai.event.external.enums import ExternalEvents
    def warpper(name, event_context):
      return name == event_context.name
    return {
      Mirai: lambda k: True,
      GroupMessage: lambda k: k.name == "GroupMessage",
      FriendMessage: lambda k: k.name =="FriendMessage",
      MessageChain: lambda k: k.name in MessageTypes,
      components.Source: lambda k: k.name in MessageTypes,
      Group: lambda k: k.name == "GroupMessage",
      Friend: lambda k: k.name =="FriendMessage",
      Member: lambda k: k.name == "GroupMessage",
      "Sender": lambda k: k.name in MessageTypes,
      "Type": lambda k: k.name,
      **({
        event_class.value: partial(warpper, copy.copy(event_name))
        for event_name, event_class in \
          ExternalEvents.__members__.items()
      })
    }

  def checkEventBodyAnnotations(self):
    event_bodys: Dict[Callable, List[str]] = {}
    for event_name in self.event:
      event_body_list = self.event[event_name]
      for i in event_body_list:
        if not event_bodys.get(i['func']):
          event_bodys[i['func']] = [event_name]
        else:
          event_bodys[i['func']].append(event_name)
    
    restraint_mapping = self.getRestraintMapping()
    
    for func in event_bodys:
      whileList = self.signature_getter(func)
      for param_name, func_item in func.__annotations__.items():
        if param_name not in whileList:
          for event_name in event_bodys[func]:
            try:
              if not (restraint_mapping[func_item](
                  type("checkMockType", (object,), {
                    "name": event_name
                  })
                )
              ):
                raise ValueError(f"error in annotations checker: {func}.{func_item}: {event_name}")
            except KeyError:
              raise ValueError(f"error in annotations checker: {func}.{func_item} is invaild.")
            except ValueError:
              raise

  def getFuncRegisteredEvents(self, callable_target: Callable):
    event_bodys: Dict[Callable, List[str]] = {}
    for event_name in self.event:
      event_body_list = sum([list(i.values()) for i in self.event[event_name]], [])
      for i in event_body_list:
        if not event_bodys.get(i['func']):
          event_bodys[i['func']] = [event_name]
        else:
          event_bodys[i['func']].append(event_name)
    return event_bodys.get(callable_target)

  def checkFuncAnnotations(self, callable_target: Callable):
    restraint_mapping = self.getRestraintMapping()
    whileList = self.signature_getter(callable_target)
    registered_events = self.getFuncRegisteredEvents(callable_target)
    for param_name, func_item in callable_target.__annotations__.items():
      if param_name not in whileList:
        if not registered_events:
          raise ValueError(f"error in annotations checker: {callable_target} is invaild.")
        for event_name in registered_events:
          try:
            if not (restraint_mapping[func_item](
                type("checkMockType", (object,), {
                  "name": event_name
                })
              )
            ):
              raise ValueError(f"error in annotations checker: {callable_target}.{func_item}: {event_name}")
          except KeyError:
            raise ValueError(f"error in annotations checker: {callable_target}.{func_item} is invaild.")
          except ValueError:
            raise

  def checkDependencies(self, depend_target: Depend):
    signature_mapping = self.signature_checker(depend_target.func)
    for k, v in signature_mapping.items():
      if type(v) == Depend:
        self.checkEventBodyAnnotations()
        self.checkDependencies(v)

  def checkEventDependencies(self):
    for event_name, event_bodys in self.event.items():
      for i in event_bodys:
        for depend in i['dependencies']:
          if type(depend) != Depend:
            raise TypeError(f"error in dependencies checker: {i['func']}: {event_name}")
          else:
            self.checkDependencies(depend)

  def exception_handler(self, exception_class=None):
    from .event.builtins import UnexpectedException
    from mirai.event.external.enums import ExternalEvents
    def receiver_warpper(
      func: Callable[[Union[FriendMessage, GroupMessage], "Session"], Awaitable[Any]]
    ):
      event_name = "UnexpectedException"

      if not inspect.iscoroutinefunction(func):
        raise TypeError("event body must be a coroutine function.")
    
      async def func_warpper_inout(context: UnexpectedException, *args, **kwargs):
        if type(context.error) == exception_class:
          return await func(context, *args, **kwargs)

      func_warpper_inout.__annotations__.update(func.__annotations__)

      protocol = {
        "func": func_warpper_inout,
        "dependencies": [],
        "middlewares": []
      }
      
      if event_name not in self.event:
        self.event[event_name] = [protocol]
      else:
        self.event[event_name].append(protocol)

      if exception_class:
        if exception_class not in self.listening_exceptions:
          self.listening_exceptions.append(exception_class)
      return func
    return receiver_warpper

  def gen_event_anno(self):
    from mirai.event.external.enums import ExternalEvents
    result = {}
    for event_name, event_class in ExternalEvents.__members__.items():
      def warpper(name, event_context):
        if name != event_context.name:
          raise ValueError("cannot look up a non-listened event.")
        return event_context.body
      result[event_class.value] = partial(warpper, copy.copy(event_name))
    return result

  def get_annotations_mapping(self):
    return {
      Mirai: lambda k: self,
      GroupMessage: lambda k: k.body \
        if k.name == "GroupMessage" else\
          raiser(ValueError("you cannot setting a unbind argument.")),
      FriendMessage: lambda k: k.body \
        if k.name == "FriendMessage" else\
          raiser(ValueError("you cannot setting a unbind argument.")),
      MessageChain: lambda k: k.body.messageChain\
        if k.name in MessageTypes else\
          raiser(ValueError("MessageChain is not enable in this type of event.")),
      components.Source: lambda k: k.body.messageChain.getSource()\
        if k.name in MessageTypes else\
          raiser(TypeError("Source is not enable in this type of event.")),
      Group: lambda k: k.body.sender.group\
        if k.name == "GroupMessage" else\
          raiser(ValueError("Group is not enable in this type of event.")),
      Friend: lambda k: k.body.sender\
        if k.name == "FriendMessage" else\
          raiser(ValueError("Friend is not enable in this type of event.")),
      Member: lambda k: k.body.sender\
        if k.name == "GroupMessage" else\
          raiser(ValueError("Group is not enable in this type of event.")),
      "Sender": lambda k: k.body.sender\
        if k.name in MessageTypes else\
          raiser(ValueError("Sender is not enable in this type of event.")),
      "Type": lambda k: k.name,
      **self.gen_event_anno()
    }

  def getEventCurrentName(self, event_value):
    from .event.builtins import UnexpectedException
    from mirai.event.external.enums import ExternalEvents
    if inspect.isclass(event_value) and issubclass(event_value, ExternalEvent): # subclass
      return event_value.__name__
    elif isinstance(event_value, ( # normal class
      UnexpectedException,
      GroupMessage,
      FriendMessage
    )):
      return event_value.__name__
    elif event_value in [ # message
      GroupMessage,
      FriendMessage
    ]:
      return event_value.__name__
    elif isinstance(event_value, ( # enum
      MessageItemType,
      ExternalEvents
    )):
      return event_value.name
    else:
      return event_value

  @property
  def registeredEventNames(self):
    return [self.getEventCurrentName(i) for i in self.event.keys()]

  def subroutine(self, func: Callable[["Mirai"], Any]):
    from .event.builtins import UnexpectedException
    async def warpper(app: "Mirai"):
      try:
        return await func(app)
      except Exception as e:
        await self.queue.put(InternalEvent(
          name="UnexpectedException",
          body=UnexpectedException(
            error=e,
            event=None,
            application=self
          )
        ))
    self.subroutines.append(warpper)
    return func

  async def checkWebsocket(self, force=False):
    return (await self.getConfig())["enableWebsocket"]

  @staticmethod
  async def run_func(func, *args, **kwargs):
    if inspect.iscoroutinefunction(func):
      await func(*args, **kwargs)
    else:
      func(*args, **kwargs)

  def onStage(self, stage_name):
    def warpper(func):
      self.lifecycle.setdefault(stage_name, [])
      self.lifecycle[stage_name].append(func)
      return func
    return warpper

  def include_others(self, *args: List["Mirai"]):
    for other in args:
      for event_name, items in other.event.items():
        if event_name in self.event:
          self.event[event_name] += items
        else:
          self.event[event_name] = items.copy()
      self.subroutines = other.subroutines
      for life_name, items in other.lifecycle:
        self.lifecycle[life_name] += items
      self.listening_exceptions += other.listening_exceptions

  def run(self, loop=None, no_polling=False, no_forever=False):
    self.checkEventBodyAnnotations()
    self.checkEventDependencies()

    loop = loop or asyncio.get_event_loop()
    self.queue = queue = asyncio.Queue(loop=loop)
    exit_signal = False
    loop.run_until_complete(self.enable_session())
    if not no_polling:
      # check ws status
      if self.useWebsocket:
        SessionLogger.info("event receive method: websocket")
      else:
        SessionLogger.info("event receive method: http polling")

      result = loop.run_until_complete(self.checkWebsocket())
      if not result: # we can use http, not ws.
        # should use http, but we can change it.
        if self.useWebsocket:
          SessionLogger.warning("catched wrong config: enableWebsocket=false, we will modify it on launch.")
          loop.run_until_complete(self.setConfig(enableWebsocket=True))
          loop.create_task(self.ws_event_receiver(lambda: exit_signal, queue))
        else:
          loop.create_task(self.message_polling(lambda: exit_signal, queue))
      else: # we can use websocket, it's fine
        if self.useWebsocket:
          loop.create_task(self.ws_event_receiver(lambda: exit_signal, queue))
        else:
          SessionLogger.warning("catched wrong config: enableWebsocket=true, we will modify it on launch.")
          loop.run_until_complete(self.setConfig(enableWebsocket=False))
          loop.create_task(self.message_polling(lambda: exit_signal, queue))
      loop.create_task(self.event_runner(lambda: exit_signal, queue))
    
    if not no_forever:
      for i in self.subroutines:
        loop.create_task(i(self))

    try:
      for start_callable in self.lifecycle['start']:
        loop.run_until_complete(self.run_func(start_callable, self))

      for around_callable in self.lifecycle['around']:
        loop.run_until_complete(self.run_func(around_callable, self))

      loop.run_forever()
    except KeyboardInterrupt:
      SessionLogger.info("catched Ctrl-C, exiting..")
    except Exception as e:
      traceback.print_exc()
    finally:
      for around_callable in self.lifecycle['around']:
        loop.run_until_complete(self.run_func(around_callable, self))

      for end_callable in self.lifecycle['end']:
        loop.run_until_complete(self.run_func(end_callable, self))

      loop.run_until_complete(self.release())


from enum import Enum
from pydantic import BaseModel

__all__ = [
    "MessageComponentTypes",
    "BaseMessageComponent"
]

class MessageComponentTypes(Enum):
    Source = "Source"
    Plain = "Plain"
    Face = "Face"
    At = "At"
    AtAll = "AtAll"
    Image = "Image"
    Quote = "Quote"
    Xml = "Xml"
    Json = "Json"
    App = "App"
    Unknown = "Unknown"

class BaseMessageComponent(BaseModel):
    type: MessageComponentTypes

    def toString(self):
        return self.__repr__()


""" python-mirai 自带的一些小型依赖注入设施.

各个函数皆返回 mirai.Depend 实例, 不需要进一步的包装.

"""

from mirai.depend import Depend
from mirai import MessageChain, Cancelled, Image, Mirai, At, Group
import re
from typing import List, Union

def RegexMatch(pattern):
    async def regex_depend_wrapper(message: MessageChain):
        if not re.match(pattern, message.toString()):
            raise Cancelled
    return Depend(regex_depend_wrapper)

def StartsWith(string):
    async def startswith_wrapper(message: MessageChain):
        if not message.toString().startswith(string):
            raise Cancelled
    return Depend(startswith_wrapper)

def WithPhoto(num=1):
    "断言消息中图片的数量"
    async def photo_wrapper(message: MessageChain):
        if len(message.getAllofComponent(Image)) < num:
            raise Cancelled
    return Depend(photo_wrapper)

def AssertAt(qq=None):
    "断言是否at了某人, 如果没有给出则断言是否at了机器人"
    async def at_wrapper(app: Mirai, message: MessageChain):
        at_set: List[At] = message.getAllofComponent(At)
        qq = qq or app.qq
        if at_set:
            for at in at_set:
                if at.target == qq:
                    return
        else:
            raise Cancelled
    return Depend(at_wrapper)

def GroupsRestraint(*groups: List[Union[Group, int]]):
    "断言事件是否发生在某个群内"
    async def gr_wrapper(app: Mirai, group: Group):
        groups = [group if isinstance(group, int) else group.id for group in groups]
        if group.id not in groups:
            raise Cancelled
    return Depend(gr_wrapper)


from typing import (
    List, Callable
)
import inspect

class Command:
    match_string: str # 主匹配
    aliases: List[str] = [] # 一堆匹配
    actions: List[Callable] = [] # 类似js的promise链
    priority: int = 0 # 优先级, 越大越会捕捉.

    def __init__(self,
        match_string: str,
        aliases: List[str] = []
    ):
        self.match_string = match_string
        self.aliases = aliases
    
    def action(self, func):
        if not callable(func):
            raise TypeError("an action must be callable.")
        self.actions.append(func)
        return func

    @staticmethod
    async def runner(func, *args, **kwargs):
        if inspect.iscoroutinefunction(func):
            return await func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    def getFristAction(self):
        if self.actions:
            return self.actions[0]

    async def run(self, *args, **kwargs):
        if self.actions:
            result = self.runner(self.actions[0], *args, **kwargs)
            for i in self.actions[1:]:
                result = self.runner(i, *args, **kwargs)
            return result


import mirai.logger
from mirai.misc import (
    ImageType
)
from mirai.face import QQFaces
from mirai.exceptions import NetworkError, Cancelled
from mirai.depend import Depend

import mirai.event.message.base
from mirai.event.message.components import (
    At,
    Plain,
    Source,
    AtAll,
    Face,
    Quote,
    Json as JsonMessage,
    Xml as XmlMessage,
    App as LightApp,
    Image
)
from mirai.event.message.chain import (
    MessageChain
)
from mirai.event.message.models import (
    GroupMessage,
    FriendMessage,
    BotMessage
)

from mirai.event import (
    InternalEvent,
    ExternalEvent
)

from mirai.event.external import (
    BotOnlineEvent,
    BotOfflineEventActive,
    BotOfflineEventForce,
    BotOfflineEventDropped,
    BotReloginEvent,
    BotGroupPermissionChangeEvent,
    BotMuteEvent,
    BotUnmuteEvent,
    BotJoinGroupEvent,

    GroupNameChangeEvent,
    GroupEntranceAnnouncementChangeEvent,
    GroupMuteAllEvent,

    # 群设置被修改事件
    GroupAllowAnonymousChatEvent,
    GroupAllowConfessTalkEvent,
    GroupAllowMemberInviteEvent,

    # 群事件(被 Bot 监听到的, 为"被动事件", 其中 Bot 身份为第三方.)
    MemberJoinEvent,
    MemberLeaveEventKick,
    MemberLeaveEventQuit,
    MemberCardChangeEvent,
    MemberSpecialTitleChangeEvent,
    MemberPermissionChangeEvent,
    MemberMuteEvent,
    MemberUnmuteEvent
)


from mirai.entities.friend import (
    Friend
)
from mirai.entities.group import (
    Group,
    Member,
    MemberChangeableSetting,
    Permission,
    GroupSetting
)

import mirai.network
import mirai.protocol

from mirai.application import Mirai
from mirai.event.builtins import (
    UnexpectedException
)
from mirai.event.external.enums import ExternalEvents


from uuid import UUID
from pydantic import BaseModel
from typing import Optional

class Friend(BaseModel):
    id: int
    nickname: Optional[str]
    remark: Optional[str]

    def __repr__(self):
        return f"<Friend id={self.id} nickname='{self.nickname}' remark='{self.remark}'>"

    def getAvatarUrl(self) -> str:
        return f'http://q4.qlogo.cn/g?b=qq&nk={self.id}&s=140'

    


from . import *
from ..builtins import UnexpectedException
from enum import Enum

class ExternalEvents(Enum):
    BotOnlineEvent = BotOnlineEvent
    BotOfflineEventActive = BotOfflineEventActive
    BotOfflineEventForce = BotOfflineEventForce
    BotOfflineEventDropped = BotOfflineEventDropped
    BotReloginEvent = BotReloginEvent
    BotGroupPermissionChangeEvent = BotGroupPermissionChangeEvent
    BotMuteEvent = BotMuteEvent
    BotUnmuteEvent = BotUnmuteEvent
    BotJoinGroupEvent = BotJoinGroupEvent

    GroupRecallEvent = GroupRecallEvent
    FriendRecallEvent = FriendRecallEvent

    GroupNameChangeEvent = GroupNameChangeEvent
    GroupEntranceAnnouncementChangeEvent = GroupEntranceAnnouncementChangeEvent
    GroupMuteAllEvent = GroupMuteAllEvent

    # 群设置被修改事件
    GroupAllowAnonymousChatEvent = GroupAllowAnonymousChatEvent # 群设置 是否允许匿名聊天 被修改
    GroupAllowConfessTalkEvent = GroupAllowConfessTalkEvent # 坦白说
    GroupAllowMemberInviteEvent = GroupAllowMemberInviteEvent # 邀请进群

    # 群事件(被 Bot 监听到的, 为被动事件, 其中 Bot 身份为第三方.)
    MemberJoinEvent = MemberJoinEvent
    MemberLeaveEventKick = MemberLeaveEventKick
    MemberLeaveEventQuit = MemberLeaveEventQuit
    MemberCardChangeEvent = MemberCardChangeEvent
    MemberSpecialTitleChangeEvent = MemberSpecialTitleChangeEvent
    MemberPermissionChangeEvent = MemberPermissionChangeEvent
    MemberMuteEvent = MemberMuteEvent
    MemberUnmuteEvent = MemberUnmuteEvent

    UnexpectedException = UnexpectedException


from . import InternalEvent
from pydantic import BaseModel
from mirai import Mirai

class UnexpectedException(BaseModel):
    error: Exception
    event: InternalEvent
    application: Mirai

    class Config:
        arbitrary_types_allowed = True


from pydantic import BaseModel
from mirai.event import ExternalEvent
from mirai.event.enums import ExternalEventTypes as EventType
from mirai.entities.group import Permission, Group, Member
from mirai.entities.friend import Friend
import typing as T
from datetime import datetime

class BotOnlineEvent(ExternalEvent):
    type: EventType = EventType.BotOnlineEvent
    qq: int

class BotOfflineEventActive(ExternalEvent):
    type: EventType = EventType.BotOfflineEventActive
    qq: int

class BotOfflineEventForce(ExternalEvent):
    type: EventType = EventType.BotOfflineEventForce
    qq: int

class BotOfflineEventDropped(ExternalEvent):
    type: EventType = EventType.BotOfflineEventDropped
    qq: int

class BotReloginEvent(ExternalEvent):
    type: EventType = EventType.BotReloginEvent
    qq: int

class BotGroupPermissionChangeEvent(ExternalEvent):
    type: EventType = EventType.BotGroupPermissionChangeEvent
    origin: Permission
    current: Permission
    group: Group

class BotMuteEvent(ExternalEvent):
    type: EventType = EventType.BotMuteEvent
    durationSeconds: int
    operator: T.Optional[Member]

class BotUnmuteEvent(ExternalEvent):
    type: EventType = EventType.BotUnmuteEvent
    operator: T.Optional[Member]

class BotJoinGroupEvent(ExternalEvent):
    type: EventType = EventType.BotJoinGroupEvent
    group: Group

class GroupRecallEvent(ExternalEvent):
    type: EventType = EventType.GroupRecallEvent
    authorId: int
    messageId: int
    time: datetime
    group: Group
    operator: Member

class FriendRecallEvent(ExternalEvent):
    type: EventType = EventType.FriendRecallEvent
    authorId: int
    messageId: int
    time: int
    operator: int

class GroupNameChangeEvent(ExternalEvent):
    type: EventType = EventType.GroupNameChangeEvent
    origin: str
    current: str
    group: Group
    isByBot: bool

class GroupEntranceAnnouncementChangeEvent(ExternalEvent):
    type: EventType = EventType.GroupEntranceAnnouncementChangeEvent
    origin: str
    current: str
    group: Group
    operator: T.Optional[Member]

class GroupMuteAllEvent(ExternalEvent):
    type: EventType = EventType.GroupMuteAllEvent
    origin: bool
    current: bool
    group: Group
    operator: T.Optional[Member]

class GroupAllowAnonymousChatEvent(ExternalEvent):
    type: EventType = EventType.GroupAllowAnonymousChatEvent
    origin: bool
    current: bool
    group: Group
    operator: T.Optional[Member]

class GroupAllowConfessTalkEvent(ExternalEvent):
    type: EventType = EventType.GroupAllowAnonymousChatEvent
    origin: bool
    current: bool
    group: Group
    isByBot: bool

class GroupAllowMemberInviteEvent(ExternalEvent):
    type: EventType = EventType.GroupAllowMemberInviteEvent
    origin: bool
    current: bool
    group: Group
    operator: T.Optional[Member]

class MemberJoinEvent(ExternalEvent):
    type: EventType = EventType.MemberJoinEvent
    member: Member

class MemberLeaveEventKick(ExternalEvent):
    type: EventType = EventType.MemberLeaveEventKick
    member: Member
    operator: T.Optional[Member]

class MemberLeaveEventQuit(ExternalEvent):
    type: EventType = EventType.MemberLeaveEventQuit
    member: Member

class MemberCardChangeEvent(ExternalEvent):
    type: EventType = EventType.MemberCardChangeEvent
    origin: str
    current: str
    member: Member
    operator: T.Optional[Member]

class MemberSpecialTitleChangeEvent(ExternalEvent):
    type: EventType = EventType.MemberSpecialTitleChangeEvent
    origin: str
    current: str
    member: Member

class MemberPermissionChangeEvent(ExternalEvent):
    type: EventType = EventType.MemberPermissionChangeEvent
    origin: str
    current: str
    member: Member

class MemberMuteEvent(ExternalEvent):
    type: EventType = EventType.MemberMuteEvent
    durationSeconds: int
    member: Member
    operator: T.Optional[Member]

class MemberUnmuteEvent(ExternalEvent):
    type: EventType = EventType.MemberUnmuteEvent
    member: Member
    operator: T.Optional[Member]